from enum import Enum


class CompetitionType(Enum):
    COMPETITORS_MENTION = "Competitors Mention"
    MARKET_MANIPULATION = "Market Manipulation"
    DISCREDITATION = "Discreditation"
    CONFIDENTIAL_STRATEGIES = "Confidential Strategies"
