from enum import Enum


class IllegalActivityType(Enum):
    WEAPONS = "Weapons"
    ILLEGAL_DRUGS = "Illegal Drugs"
    VIOLENT_CRIME = "Violent Crimes"
    NON_VIOLENT_CRIME = "Non Violent Crimes"
    SEX_CRIME = "Sex Crimes"
    CYBERCRIME = "Cybercrime"
    CHILD_EXPLOITATION = "Child Exploitation"
