from enum import Enum


class IntellectualPropertyType(Enum):
    IMITATION = "Imitation"
    COPYRIGHT_VIOLATIONS = "Copyright Violations"
    TRADEMARK_INFRINGEMENT = "Trademark Infringement"
    PATENT_DISCLOSURE = "Patent Disclosure"
