from enum import Enum


class PersonalSafetyType(Enum):
    BULLYING = "Bullying"
    SELF_HARM = "Self Harm"
    UNSAFE_PRACTICES = "Unsafe Practices"
    DANGEROUS_CHALLENGES = "Dangerous Challenges"
    STALKING = "Stalking"
